package pipeline

import (
	"testing"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
	"github.com/stretchr/testify/assert"
	"gopkg.in/yaml.v2"
)

func TestStringSetEqual(t *testing.T) {
	for _, tt := range []struct {
		name    string
		s1      []string
		s2      []string
		expects bool
	}{
		{"empty", nil, nil, true},
		{"same one item", []string{"i1"}, []string{"i1"}, true},
		{"different one item", []string{"i1"}, []string{"i2"}, false},
		{"same 2 items, same order", []string{"i1", "i2"}, []string{"i1", "i2"}, true},
		{"same 2 items, different order", []string{"i1", "i2"}, []string{"i2", "i1"}, true},
	} {
		t.Run(tt.name, func(t *testing.T) {
			assert.Equal(t, tt.expects, stringSetEqual(tt.s1, tt.s2))
		})
	}
}

func TestStringSetOverlap(t *testing.T) {
	for _, tt := range []struct {
		name    string
		s1      []string
		s2      []string
		expects bool
	}{
		{"empty", nil, nil, false},
		{"same one item", []string{"i1"}, []string{"i1"}, true},
		{"different one item", []string{"i1"}, []string{"i2"}, false},
		{"same 2 items, same order", []string{"i1", "i2"}, []string{"i1", "i2"}, true},
		{"same 2 items, different order", []string{"i1", "i2"}, []string{"i2", "i1"}, true},
		{"1 common item", []string{"i1", "i2", "i3", "i4"}, []string{"i3", "i5", "i6"}, true},
	} {
		t.Run(tt.name, func(t *testing.T) {
			assert.Equal(t, tt.expects, stringSetOverlap(tt.s1, tt.s2))
		})
	}
}

func TestStringSetContains(t *testing.T) {
	for _, tt := range []struct {
		name    string
		set     []string
		s       string
		expects bool
	}{
		{"empty", nil, "rabbit", false},
		{"one item matching", []string{"i1"}, "i1", true},
		{"one item not matching", []string{"i1"}, "i2", false},
		{"2 items, one matching", []string{"i1", "i2"}, "i2", true},
		{"2 items, none matching", []string{"i1", "i2"}, "i3", false},
	} {
		t.Run(tt.name, func(t *testing.T) {
			assert.Equal(t, tt.expects, stringSetContains(tt.set, tt.s))
		})
	}
}

func TestUUIDSetContains(t *testing.T) {
	var (
		u1 = api.MustUUIDFromString("93fd5400-1d8a-11e9-9c29-4f8fa1202874")
		u2 = api.MustUUIDFromString("94b36966-1d8a-11e9-8993-6fb5e8bbc3b5")
		u3 = api.MustUUIDFromString("95c6ff84-1d8a-11e9-b6c2-4fb5207ffb8a")
	)
	for _, tt := range []struct {
		name    string
		set     []api.UUID
		s       api.UUID
		expects bool
	}{
		{"empty", nil, u1, false},
		{"one item matching", []api.UUID{u1}, u1, true},
		{"one item not matching", []api.UUID{u1}, u2, false},
		{"2 items, one matching", []api.UUID{u1, u2}, u2, true},
		{"2 items, none matching", []api.UUID{u1, u2}, u3, false},
	} {
		t.Run(tt.name, func(t *testing.T) {
			assert.Equal(t, tt.expects, uuidSetContains(tt.set, tt.s))
		})
	}
}

func TestPipelineMatch(t *testing.T) {
	var (
		uuid1                  = api.MustUUIDFromString("74754054-1d88-11e9-93c3-337ecc46b938")
		uuid2                  = api.MustUUIDFromString("7de696d8-1d88-11e9-a266-6bf1e33eec2c")
		simplePipelineForRole1 = `
graph:
  nodes:
  - id: source
    type: emitter
    roles: [role1]
    sourcematch:
      eventtypes:
      - event1
    outputs: [default]
`
		simplePipelineForActor1 = `
graph:
  nodes:
  - id: source
    type: emitter
    actors: [actor1]
    sourcematch:
      eventtypes:
      - event1
    outputs: [default]
`
		simplePipelineForActorID1 = `
graph:
  nodes:
  - id: source
    type: emitter
    actorids: [74754054-1d88-11e9-93c3-337ecc46b938]
    sourcematch:
      eventtypes:
      - event1
    outputs: [default]
`
		actor1 = api.Actor{
			ID:    uuid1,
			Name:  "actor1",
			Kind:  api.Actor_EMITTER,
			Roles: []string{"role1"},
		}
		actor2 = api.Actor{
			ID:    uuid2,
			Name:  "actor2",
			Kind:  api.Actor_EMITTER,
			Roles: []string{"role2"},
		}
	)

	for _, tt := range []struct {
		name       string
		pipeline   string
		actor      api.Actor
		eventtypes []string
		expected   *storage.NodeOutputRef
	}{
		{"simplematch", simplePipelineForRole1,
			actor1, []string{"event1"},
			&storage.NodeOutputRef{NodeID: "source", Output: "default"}},
		{"wrongevent", simplePipelineForRole1,
			actor1, []string{"event2"},
			nil},
		{"wrongrole", simplePipelineForRole1,
			actor2, []string{"event1"},
			nil},
		{"matchactor", simplePipelineForActor1,
			actor1, []string{"event1"},
			&storage.NodeOutputRef{NodeID: "source", Output: "default"}},
		{"matchactor_wrongactor", simplePipelineForActor1,
			actor2, []string{"event1"},
			nil},
		{"matchactorid", simplePipelineForActorID1,
			actor1, []string{"event1"},
			&storage.NodeOutputRef{NodeID: "source", Output: "default"}},
		{"matchactorid_wrongactor", simplePipelineForActorID1,
			actor2, []string{"event1"},
			nil},
	} {
		t.Run(tt.name, func(t *testing.T) {
			var sPipe storage.Pipeline
			if err := yaml.Unmarshal([]byte(tt.pipeline), &sPipe); err != nil {
				panic(err)
			}
			pipe := Pipeline{sPipe}
			noderef, match := pipe.Match(tt.actor, tt.eventtypes)
			if tt.expected == nil {
				assert.False(t, match)
			} else {
				assert.True(t, match)
				assert.Equal(t, tt.expected.String(), noderef.String())
			}
		})
	}
}
